/*
 * HSP.h
 *
 *  Created on: Feb 19, 2010
 *      Author: kohany
 */

#ifndef HSP_H_
#define HSP_H_

#include "Fragment.h"
#include <stdexcept>
#include <sstream>

using namespace std;

namespace bio {

enum CMPVAL {
   TOPBEG, TOPEND, BOTBEG, BOTEND, SCORE, SIM, POS
};

class HSP {
   Fragment top;
   Fragment bot;
   float sim;
   float pos;
   float score;
   static CMPVAL cmpby;
public:
   HSP();
   HSP(const HSP& other);
   virtual ~HSP();
   HSP& operator =(const HSP& other);
   const Fragment& getTop() const {
      return top;
   }
   const Fragment& getBot() const {
      return bot;
   }
   Fragment& getTop() {
      return top;
   }
   Fragment& getBot() {
      return bot;
   }
   const float getSim() const {
      return sim;
   }
   const float getPos() const {
      return pos;
   }
   const float getScore() const {
      return score;
   }
   void setSim(float sim) {
      this->sim = sim;
   }
   void setPos(float pos) {
      this->pos = pos;
   }
   void setScore(float score) {
      this->score = score;
   }
   const bool isDirect() const {
      return !(getTop().isDirect() ^ getBot().isDirect());
   }
   virtual HSP& operator +=(const HSP& other);
   virtual HSP& operator -=(const HSP& other);
   pair<HSP, HSP> subsplit(const HSP& other) const;
   const HSP operator +(const HSP& other) {
      HSP res(*this);
      return (res += other);
   }
   const HSP operator -(const HSP& other) {
      HSP res(*this);
      return (res -= other);
   }
   static void setCmpBy(CMPVAL cmpby) {
      HSP::cmpby = cmpby;
   }
   virtual const double getCmpVal() const;
   const bool operator >(const HSP& other) const;
   const bool operator ==(const HSP& other) const;
   const bool operator <(const HSP& other) const;
   const bool operator <=(const HSP& other) const;
   const bool operator >=(const HSP& other) const;
   virtual HSP& read(istream& is);
   virtual void write(ostream& os) const;
};

istream& operator >>(istream& is, HSP& hsp);
ostream& operator <<(ostream& os, const HSP& hsp);

/**
 * This is a utility class used for reading HSP from input stream that does 2 things:
 * 1. Avoids multiple constructions of istringstream as in operator >> (ostream&, HSP&)
 * 2. Reduces number of copying of HSP objects by reading directly into target HSP.
 */
class HSPReader {
   string line;
   istringstream iss;
public:
   HSPReader() {
      iss.exceptions(istream::failbit | istream::badbit);
   }
   void readHSP(istream& is, HSP& hsp) {
      getline(is, line);
      iss.str(line);
      iss.clear();
      try {
         hsp.read(iss);
      }
      catch(const istream::failure& err) {
         iss.clear( iss.rdstate() & (~istream::failbit));
         throw(istream::failure(string("ERROR: Failed to parse line:\nERROR: '")+iss.str()+"'"));
      }
   }
};

/**
 * Checks if two HSPs can be joined: have the same orientation, same sequence names,
 * are within maximum allowed distance maxDist from each other.
 */
bool areJoinable(const HSP& hsp1, const HSP& hsp2, float proxiRatio = 0.15, int maxDist = MAXDIST);
double interpolateScore(double score1, double ctr1, double score2, double ctr2, double shared);
void testHSP();
};
#endif /* HSP_H_ */
