/*
 * ScoreMatrix.h
 *
 *  Created on: Aug 13, 2009
 *      Author: kohany
 */

#ifndef SCOREMATRIX_H_
#define SCOREMATRIX_H_

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <stdexcept>
#include <string.h>

using namespace std;
namespace bio {
   enum SEQTYPE {
      SEQTYPE_UNKNOWN, NA, AA
   };
   /**
    * Template parameter - a type for score values.
    */
   template<typename T>
   class ScoreMatrix {
   public:
      typedef short index_t;
      typedef char char_t;
      typedef T score_t;
      typedef vector<score_t> scorev_t;
      typedef typename scorev_t::iterator iterator;
      typedef typename scorev_t::const_iterator const_iterator;
      typedef typename scorev_t::reverse_iterator reverse_iterator;
      typedef typename scorev_t::const_reverse_iterator const_reverse_iterator;
   private:
      scorev_t matrix;
      index_t char2index[255];
      char_t index2char[255];
      char alphabethSize;
      score_t gapIniPenalty;
      score_t gapExtPenalty;
      score_t gapMatchScore; // Used to score GAP-to-GAP pair.
      string& trim(string& s, char c = ' ');
      string& removeComments(string& line);
   public:
      static const char DNAALNMATRIX[];
      static const char DNACONMATRIX[];
      static const char PROTALNMATRIX[];
      static const char GAPCHAR;
      static const char STOPCODON;
      static const char LPADDING;
      static const char RPADDING;
      ScoreMatrix();
      ScoreMatrix(const ScoreMatrix& other);
      explicit ScoreMatrix(const char* matrix, bool fromFile = true);
      virtual ~ScoreMatrix();
      const_iterator begin() const {
         return matrix.begin();
      }
      const_iterator end() const {
         return matrix.end();
      }
      const_iterator getRow(size_t index) const {
         return matrix.begin() + index * alphabethSize;
      }
      ScoreMatrix& assign(const ScoreMatrix& other);
      ScoreMatrix& operator =(const ScoreMatrix& other);
      bool loadMatrixFromString(const char* defaultMatrix = ScoreMatrix::DNAALNMATRIX);
      bool parse(const char* file);
      bool parse(istream& is);
      void print(ostream& os) const;
      score_t getScore(char c1, char c2) const;
      string getAlphabeth() const;
      score_t getMaxScore() const;
      score_t getMinScore() const;
      score_t getGapIniPenalty() const {
         return gapIniPenalty;
      }
      score_t getGapExtPenalty() const {
         return gapExtPenalty;
      }
      score_t getGapMatchScore() const {
         return gapMatchScore;
      }
      void setGapIniPenalty(score_t penalty) {
         gapIniPenalty = penalty;
      }
      void setGapExtPenalty(score_t penalty) {
         gapExtPenalty = penalty;
      }
      void setGapMatchScore(score_t score) {
         gapMatchScore = score;
      }

      const scorev_t& getMatrix() const {
         return matrix;
      }
      const index_t* getChar2Index() const {
         return char2index;
      }
      const char_t* getIndex2Char() const {
         return index2char;
      }
      index_t c2i(char_t c) const {
         return char2index[c];
      }
      char_t i2c(index_t i) const {
         return index2char[i];
      }
      template <typename InputIterator, typename OutputIterator>
      void c2i(InputIterator beg, InputIterator end, OutputIterator out) const;
      template <typename InputIterator, typename OutputIterator>
      void i2c(InputIterator beg, InputIterator end, OutputIterator out) const;
      char getAlphabethSize() const {
         return alphabethSize;
      }
      SEQTYPE getType() const {
         if(alphabethSize > 20) return AA;
         return NA;
      }
   };

   template<typename T>
   const char
         ScoreMatrix<T>::DNAALNMATRIX[] =
               "GAP -30 -6\n\
  A   R   G   C   Y   T   M   W   S   K   N   X\n\
  9   3  -5 -12 -12 -12   0   0  -9  -9  -1  -3\n\
  3   3   3 -12 -12 -12  -6  -6  -6  -6  -1  -3\n\
 -7   3   9 -12 -12 -12  -9  -9   0   0  -1  -3\n\
-12 -12 -12   9   3  -7   0  -9   0  -9  -1  -3\n\
-12 -12 -12   3   3   3  -6  -6  -6  -6  -1  -3\n\
-12 -12 -12  -5   3   9  -9   0  -9   0  -1  -3\n\
  0  -6  -9   0  -6  -9   0  -6  -6  -9  -1  -3\n\
  0  -6  -9  -9  -6   0  -6   0  -9  -6  -1  -3\n\
 -9  -6   0   0  -6  -9  -6  -9   0  -6  -1  -3\n\
 -9  -6   0  -9  -6   0  -9  -6  -6   0  -1  -3\n\
 -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -3\n\
 -3  -3  -3  -3  -3  -3  -3  -3  -3  -3  -3  -3";

   //This is a slightly modified copy of linupmatrix from RepeatModeler by Arian Smith and Robert Hubley
   template<typename T>
   const char
         ScoreMatrix<T>::DNACONMATRIX[] =
               "GAP -25 -5\n\
  A   R   G   C   Y   T   K   M   S   W   N   X\n\
  9   0  -8 -15 -16 -17 -13  -3 -11  -4  -5  -7\n\
  2   1   1 -15 -15 -16  -7  -6  -6  -7  -5  -7\n\
 -4   3  10 -14 -14 -15  -2  -9  -2  -9  -5  -7\n\
-15 -14 -14  10   3  -4  -9  -2  -2  -9  -5  -7\n\
-16 -15 -15   1   1   2  -6  -7  -6  -7  -5  -7\n\
-17 -16 -15  -8   0   9  -3 -13 -11  -4  -5  -7\n\
-11  -6  -2 -11  -7  -3  -2 -11  -6  -7  -5  -7\n\
 -3  -7 -11  -2  -6 -11 -11  -2  -6  -7  -5  -7\n\
 -9  -5  -2  -2  -5  -9  -5  -5  -2  -9  -5  -7\n\
 -4  -8 -11 -11  -8  -4  -8  -8 -11  -4  -5  -7\n\
 -5  -5  -5  -5  -5  -5  -5  -5  -5  -5  -1  -7\n\
 -7  -7  -7  -7  -7  -7  -7  -7  -7  -7  -7  -7";

   // BLOSUM62
   template<typename T>
   const char
         ScoreMatrix<T>::PROTALNMATRIX[] =
   "GAP -12 -2\n\
 A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V  B  Z  X  *\n\
 4 -1 -2 -2  0 -1 -1  0 -2 -1 -1 -1 -1 -2 -1  1  0 -3 -2  0 -2 -1  0 -4\n\
-1  5  0 -2 -3  1  0 -2  0 -3 -2  2 -1 -3 -2 -1 -1 -3 -2 -3 -1  0 -1 -4\n\
-2  0  6  1 -3  0  0  0  1 -3 -3  0 -2 -3 -2  1  0 -4 -2 -3  3  0 -1 -4\n\
-2 -2  1  6 -3  0  2 -1 -1 -3 -4 -1 -3 -3 -1  0 -1 -4 -3 -3  4  1 -1 -4\n\
 0 -3 -3 -3  9 -3 -4 -3 -3 -1 -1 -3 -1 -2 -3 -1 -1 -2 -2 -1 -3 -3 -2 -4\n\
-1  1  0  0 -3  5  2 -2  0 -3 -2  1  0 -3 -1  0 -1 -2 -1 -2  0  3 -1 -4\n\
-1  0  0  2 -4  2  5 -2  0 -3 -3  1 -2 -3 -1  0 -1 -3 -2 -2  1  4 -1 -4\n\
 0 -2  0 -1 -3 -2 -2  6 -2 -4 -4 -2 -3 -3 -2  0 -2 -2 -3 -3 -1 -2 -1 -4\n\
-2  0  1 -1 -3  0  0 -2  8 -3 -3 -1 -2 -1 -2 -1 -2 -2  2 -3  0  0 -1 -4\n\
-1 -3 -3 -3 -1 -3 -3 -4 -3  4  2 -3  1  0 -3 -2 -1 -3 -1  3 -3 -3 -1 -4\n\
-1 -2 -3 -4 -1 -2 -3 -4 -3  2  4 -2  2  0 -3 -2 -1 -2 -1  1 -4 -3 -1 -4\n\
-1  2  0 -1 -3  1  1 -2 -1 -3 -2  5 -1 -3 -1  0 -1 -3 -2 -2  0  1 -1 -4\n\
-1 -1 -2 -3 -1  0 -2 -3 -2  1  2 -1  5  0 -2 -1 -1 -1 -1  1 -3 -1 -1 -4\n\
-2 -3 -3 -3 -2 -3 -3 -3 -1  0  0 -3  0  6 -4 -2 -2  1  3 -1 -3 -3 -1 -4\n\
-1 -2 -2 -1 -3 -1 -1 -2 -2 -3 -3 -1 -2 -4  7 -1 -1 -4 -3 -2 -2 -1 -2 -4\n\
 1 -1  1  0 -1  0  0  0 -1 -2 -2  0 -1 -2 -1  4  1 -3 -2 -2  0  0  0 -4\n\
 0 -1  0 -1 -1 -1 -1 -2 -2 -1 -1 -1 -1 -2 -1  1  5 -2 -2  0 -1 -1  0 -4\n\
-3 -3 -4 -4 -2 -2 -3 -2 -2 -3 -2 -3 -1  1 -4 -3 -2 11  2 -3 -4 -3 -2 -4\n\
-2 -2 -2 -3 -2 -1 -2 -3  2 -1 -1 -2 -1  3 -3 -2 -2  2  7 -1 -3 -2 -1 -4\n\
 0 -3 -3 -3 -1 -2 -2 -3 -3  3  1 -2  1 -1 -2 -2  0 -3 -1  4 -3 -2 -1 -4\n\
-2 -1  3  4 -3  0  1 -1  0 -3 -4  0 -3 -3 -2  0 -1 -4 -3 -3  4  1 -1 -4\n\
-1  0  0  1 -3  3  4 -2  0 -3 -3  1 -1 -3 -1  0 -1 -3 -2 -2  1  4 -1 -4\n\
 0 -1 -1 -1 -2 -1 -1 -1 -1 -1 -1 -1 -1 -1 -2  0  0 -2 -1 -1 -1 -1 -1 -4\n\
-4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4 -4  1";

   template<typename T>
   const char ScoreMatrix<T>::GAPCHAR = '-';
   template<typename T>
   const char ScoreMatrix<T>::STOPCODON = '*';
   template<typename T>
   const char ScoreMatrix<T>::LPADDING = '<';
   template<typename T>
   const char ScoreMatrix<T>::RPADDING = '>';

   template <typename T>
   ostream& operator<<(ostream& os, const ScoreMatrix<T> mat) {
      mat.print(os);
      return os;
   }
   template <typename T>
   istream& operator>>(istream& is, ScoreMatrix<T> mat) {
      mat.parse(is);
      return is;
   }

   template<typename T>
   ScoreMatrix<T>::ScoreMatrix() :
      gapIniPenalty(0), gapExtPenalty(0), gapMatchScore(0) {
      if (!loadMatrixFromString(DNAALNMATRIX)) {
         cerr << "Failed to load default matrix" << endl;
      }
   }

   template<typename T>
   ScoreMatrix<T>::ScoreMatrix(const char* matrix, bool fromFile) :
      gapIniPenalty(0), gapExtPenalty(0), gapMatchScore(0) {
      if (fromFile) {
         if (parse(matrix)) {
            return;
         }
      } else {
         if (loadMatrixFromString(matrix)) {
            return;
         }
      }
      if (!loadMatrixFromString(DNAALNMATRIX)) {
         cerr << "Failed to load default matrix" << endl;
      }
   }

   template<typename T>
   ScoreMatrix<T>::ScoreMatrix(const ScoreMatrix& other) {
      assign(other);
   }
   template<typename T>
   ScoreMatrix<T>::~ScoreMatrix() {
   }
   template<typename T>
   ScoreMatrix<T>& ScoreMatrix<T>::operator =(const ScoreMatrix<T>& other) {
      return assign(other);
   }
   template<typename T>
   ScoreMatrix<T>& ScoreMatrix<T>::assign(const ScoreMatrix<T>& other) {
      alphabethSize = other.alphabethSize;
      matrix = other.getMatrix();
      memcpy(char2index, other.char2index, sizeof(other.char2index));
      memcpy(index2char, other.index2char, sizeof(index2char));
      gapIniPenalty = other.gapIniPenalty;
      gapExtPenalty = other.gapExtPenalty;
      gapMatchScore = other.gapMatchScore;
      return *this;
   }
   template<typename T>
   string& ScoreMatrix<T>::trim(string& s, char c) {
      size_t pos = 0;
      while (s[pos] == c) {
         pos++;
      }
      s.erase(0, pos);
      return s;
   }

   template<typename T>
   bool ScoreMatrix<T>::loadMatrixFromString(const char* defaultMatrix) {
      if (defaultMatrix != NULL) {
         string s(defaultMatrix);
         istringstream iss(s);
         return parse(iss);
      }
      return false;
   }

   template<typename T>
   bool ScoreMatrix<T>::parse(const char* file) {
      ifstream ifs;
      ifs.open(file);
      if (ifs.good()) {
         bool rv = parse(ifs);
         ifs.close();
         return rv;
      } else {
         if (file[0] != '\0') {
            cerr << "ERROR: Failed to read " << file << endl;
         }
      }
      return false;
   }

   template<typename T>
   string& ScoreMatrix<T>::removeComments(string& line) {
      /*
      size_t pos = line.find('#');
      if (pos < line.npos) {
         line.erase(pos);
      }
      */
      for(string::iterator sptr = line.begin(); sptr != line.end(); sptr++) {
         if(isgraph(*sptr)) {
            if(*sptr == '#') {
               sptr = line.erase(sptr);
            }
            break;
         }
      }
      trim(trim(line, ' '), '\t');
      return line;
   }
   /**  */
   template<typename T>
   bool ScoreMatrix<T>::parse(istream& is) {
      scorev_t newMatrix;
      typename scorev_t::iterator rowptr;
      index_t newChar2Index[255];
      char_t newIndex2Char[255];
      memset(newChar2Index, -1, sizeof(newChar2Index));
      memset(newIndex2Char, 0, sizeof(newIndex2Char));
      string line;
      int gi = gapIniPenalty;
      int gx = gapExtPenalty;
      size_t pos;
      size_t size = 0;
      size_t index = 0;
      istringstream iss;

      try {
         while (!getline(is, line).fail()) {
            char c;
            string s;
            removeComments(line);
            if (!isgraph(line[0])) {
               continue;
            }
            if (line.find("FREQS") == 0) {
               continue;
            }
            pos = line.find("GAP");
            /**
             * Reading GAP penalties if available
             */
            if (pos < line.npos) {
               s = line.substr(pos + 3);
               trim(s);
               iss.str(s);
               iss.clear();
               iss >> gi >> gx;
               if (!iss.fail()) {
                  gapIniPenalty = gi;
                  gapExtPenalty = gx;
                  gapMatchScore = -gx / 2;
               } else {
                  cerr << "ERROR: " << line << endl;
                  cerr << "ERROR: failed to read gap penalties" << endl;
               }
               continue;
            }
            //if (isalpha(line[0]) && size == 0) {
            if (size == 0) {
               iss.str(line);
               iss.clear();
               for (iss >> s; !iss.fail(); iss >> s) {
                  if (s.size() > 1) {
                     cerr << "Expected single character in score matrix column header" << endl;
                     throw logic_error("Expected single character in score matrix column header");
                  }
                  c = s[0];
                  newChar2Index[tolower(c)] = size;
                  newChar2Index[toupper(c)] = size;
                  newIndex2Char[size] = c;
                  size++;
               }
               if (size > 0) {
                  newMatrix.resize(size * size, 0);
                  rowptr = newMatrix.begin();
               }
               continue;
            }
            if (size > 0) {
               iss.str(line);
               iss.clear();
               if (isalpha(line[0])) {
                  iss >> s;
                  if (s.size() > 1) {
                     cerr << "Expected single character in score matrix row header" << endl;
                     throw logic_error("Expected single character in score matrix row header");
                  }
                  c = s[0];
                  if (newChar2Index[c] == -1) {
                     cerr << "Row index character is not present in column index" << endl;
                     throw logic_error("Row index character is not present in column index");
                  }
                  index = newChar2Index[c];
               }
               rowptr = newMatrix.begin() + size * index;
               index++;
               double v;
               size_t rowIndex = 0;
               for (iss >> v; !iss.fail(); iss >> v) {
                  rowptr[rowIndex++] = (score_t) v;
               }
               if (rowIndex < size) {
                  cerr << "Read fewer than expected values from score matrix's row: " << rowIndex << "<" << size << endl;
                  cerr << "line parsed: " << line << endl;
                  throw logic_error("Read fewer than expected values from score matrix's row");
               }
            }
         }
         alphabethSize = size;
         memcpy(char2index, newChar2Index, sizeof(newChar2Index));
         memcpy(index2char, newIndex2Char, sizeof(newIndex2Char));
         matrix = newMatrix;
         return true;
      } catch (const logic_error& exc) {
         cerr << exc.what() << " at " << __FILE__ << ":" << __LINE__ << endl;
         return false;
      }
   }

   template<typename T>
   void ScoreMatrix<T>::print(ostream& os) const {
      if (gapIniPenalty * gapExtPenalty != 0) {
         os << "GAP " << (double) gapIniPenalty << " " << (double) gapExtPenalty << endl;
      }
      for (int i = 0; i < alphabethSize; i++) {
         os << "\t" << index2char[i];
      }
      os << endl;
      for (int y = 0; y < alphabethSize; y++) {
         os << index2char[y];
         for (int x = 0; x < alphabethSize; x++) {
            os << "\t" << (double) matrix[y * alphabethSize + x];
         }
         os << endl;
      }
   }

   template<typename T>
   T ScoreMatrix<T>::getScore(char c1, char c2) const {
      if (c1 == LPADDING || c2 == LPADDING || c1 == RPADDING || c2 == RPADDING) {
         return 0;
      }
      if (c1 == GAPCHAR || c2 == GAPCHAR) {
         if (c1 != GAPCHAR || c2 != GAPCHAR) {
            return gapExtPenalty;
         }
         return gapMatchScore;
      }
      if (char2index[c1] < 0) {
         //cerr << "ERROR: Scores undefined for specified symbol c1='" << c1 << "' [" << (int) c1 << "]" << endl;
         throw logic_error("ERROR: Score undefined for specified symbol");
      }
      if (char2index[c2] < 0) {
         //cerr << "ERROR: Scores undefined for specified symbol c2='" << c2 << "' [" << (int) c2 << "]" << endl;
         throw logic_error("ERROR: Score undefined for specified symbol");
      }
      return matrix[char2index[c1] * alphabethSize + char2index[c2]];
   }

   template<typename T>
   string ScoreMatrix<T>::getAlphabeth() const {
      return string(index2char, alphabethSize);
   }
   template<typename T>
   T ScoreMatrix<T>::getMaxScore() const {
      if (matrix.empty()) {
         return 0;
      }
      T maxWeight = matrix[0];
      for (int i = 1; i < alphabethSize; i++) {
         if (maxWeight < matrix[i * alphabethSize + i]) {
            maxWeight = matrix[i * alphabethSize + i];
         }
      }
      return maxWeight;
   }

   template<typename T>
   T ScoreMatrix<T>::getMinScore() const {
      if (matrix.empty()) {
         return 0;
      }
      T minWeight = matrix[0];
      for (int i = 1; i < alphabethSize; i++) {
         if (minWeight > matrix[i * alphabethSize + i]) {
            minWeight = matrix[i * alphabethSize + i];
         }
      }
      return minWeight;
   }
   template <typename T>
   template <typename InputIterator, typename OutputIterator>
   void ScoreMatrix<T>::c2i(InputIterator beg, InputIterator end, OutputIterator out) const {
      for(; beg != end; beg++) {
         if(c2i(*beg) == -1) {
            cerr << "WARNING: Ignoring symbol'" << *beg << "'" << endl;
         }
         else {
            *out++ = c2i(*beg);
         }
      }
   }
   template <typename T>
   template <typename InputIterator, typename OutputIterator>
   void ScoreMatrix<T>::i2c(InputIterator beg, InputIterator end, OutputIterator out) const {
      for(; beg != end; beg++) {
         if(i2c(*beg) == 0) {
            cerr << "WARNING: No symbol for index'" << *beg << "'" << endl;
         }
         else {
            *out++ = i2c(*beg);
         }
      }
   }
}
#endif /* SCOREMATRIX_H_ */
